from inewave._utils.bloco import Bloco
from typing import Any, Type, List
from inewave._utils.arquivo import Arquivo
from inewave._utils.dadosarquivo import DadosArquivo
from inewave._utils.escrita import Escrita
from inewave.newave.modelos.dger import BlocoMesInicioEstudo, BlocoNomeCaso
from inewave.newave.modelos.dger import BlocoTipoExecucao
from inewave.newave.modelos.dger import BlocoDuracaoPeriodo
from inewave.newave.modelos.dger import BlocoNumAnosEstudo
from inewave.newave.modelos.dger import BlocoMesInicioPreEstudo
from inewave.newave.modelos.dger import BlocoAnoInicioEstudo
from inewave.newave.modelos.dger import BlocoNumAnosPreEstudo
from inewave.newave.modelos.dger import BlocoNumAnosPosEstudo
from inewave.newave.modelos.dger import BlocoNumAnosPosEstudoSimFinal
from inewave.newave.modelos.dger import BlocoImprimeDados
from inewave.newave.modelos.dger import BlocoImprimeMercados
from inewave.newave.modelos.dger import BlocoImprimeEnergias
from inewave.newave.modelos.dger import BlocoImprimeModeloEstocastico
from inewave.newave.modelos.dger import BlocoImprimeSubsistema
from inewave.newave.modelos.dger import BlocoNumMaxIteracoes
from inewave.newave.modelos.dger import BlocoNumForwards
from inewave.newave.modelos.dger import BlocoNumAberturas
from inewave.newave.modelos.dger import BlocoNumSeriesSinteticas
from inewave.newave.modelos.dger import BlocoOrdemMaximaPARp
from inewave.newave.modelos.dger import BlocoAnoInicialHistorico
from inewave.newave.modelos.dger import BlocoCalculaVolInicial
from inewave.newave.modelos.dger import BlocoVolInicialSubsistema
from inewave.newave.modelos.dger import BlocoTolerancia
from inewave.newave.modelos.dger import BlocoTaxaDesconto
from inewave.newave.modelos.dger import BlocoTipoSimFinal
from inewave.newave.modelos.dger import BlocoImpressaoOperacao
from inewave.newave.modelos.dger import BlocoImpressaoConvergencia
from inewave.newave.modelos.dger import BlocoIntervaloGravar
from inewave.newave.modelos.dger import BlocoMinIteracoes
from inewave.newave.modelos.dger import BlocoRacionamentoPreventivo
from inewave.newave.modelos.dger import BlocoNumAnosManutUTE
from inewave.newave.modelos.dger import BlocoTendenciaHidrologica
from inewave.newave.modelos.dger import BlocoRestricaoItaipu
from inewave.newave.modelos.dger import BlocoBid
from inewave.newave.modelos.dger import BlocoPerdasTransmissao
from inewave.newave.modelos.dger import BlocoElNino
from inewave.newave.modelos.dger import BlocoEnso
from inewave.newave.modelos.dger import BlocoDuracaoPorPatamar
from inewave.newave.modelos.dger import BlocoOutrosUsosAgua
from inewave.newave.modelos.dger import BlocoCorrecaoDesvio
from inewave.newave.modelos.dger import BlocoCurvaAversao
from inewave.newave.modelos.dger import BlocoTipoGeracaoENA
from inewave.newave.modelos.dger import BlocoRiscoDeficit
from inewave.newave.modelos.dger import BlocoIteracaoParaSimFinal
from inewave.newave.modelos.dger import BlocoAgrupamentoLivre
from inewave.newave.modelos.dger import BlocoEqualizacaoPenalInt
from inewave.newave.modelos.dger import BlocoRepresentacaoSubmot
from inewave.newave.modelos.dger import BlocoOrdenacaoAutomatica
from inewave.newave.modelos.dger import BlocoConsideraCargaAdicional
from inewave.newave.modelos.dger import BlocoDeltaZSUP
from inewave.newave.modelos.dger import BlocoDeltaZINF
from inewave.newave.modelos.dger import BlocoDeltasConsecutivos
from inewave.newave.modelos.dger import BlocoDespachoAntecipadoGNL
from inewave.newave.modelos.dger import BlocoModifAutomaticaAdTerm
from inewave.newave.modelos.dger import BlocoGeracaoHidraulicaMin
from inewave.newave.modelos.dger import BlocoSimFinalComData
from inewave.newave.modelos.dger import BlocoGerenciamentoPLs
from inewave.newave.modelos.dger import BlocoSAR
from inewave.newave.modelos.dger import BlocoCVAR
from inewave.newave.modelos.dger import BlocoZSUPMinConvergencia
from inewave.newave.modelos.dger import BlocoDesconsideraVazaoMinima
from inewave.newave.modelos.dger import BlocoRestricoesEletricas
from inewave.newave.modelos.dger import BlocoSelecaoCortes
from inewave.newave.modelos.dger import BlocoJanelaCortes
from inewave.newave.modelos.dger import BlocoReamostragemCenarios
from inewave.newave.modelos.dger import BlocoConvergeNoZero
from inewave.newave.modelos.dger import BlocoConsultaFCF
from inewave.newave.modelos.dger import BlocoImpressaoENA
from inewave.newave.modelos.dger import BlocoImpressaoCortesAtivosSimFinal
from inewave.newave.modelos.dger import BlocoRepresentacaoAgregacao
from inewave.newave.modelos.dger import BlocoMatrizCorrelacaoEspacial
from inewave.newave.modelos.dger import BlocoDesconsideraConvEstatistica
from inewave.newave.modelos.dger import BlocoMomentoReamostragem
from inewave.newave.modelos.dger import BlocoMantemArquivosEnergias
from inewave.newave.modelos.dger import BlocoInicioTesteConvergencia
from inewave.newave.modelos.dger import BlocoSazonalizarVminT
from inewave.newave.modelos.dger import BlocoSazonalizarVmaxT
from inewave.newave.modelos.dger import BlocoSazonalizarVminP
from inewave.newave.modelos.dger import BlocoSazonalizarCfugaCmont
from inewave.newave.modelos.dger import BlocoRestricoesEmissaoGEE
from inewave.newave.modelos.dger import BlocoAfluenciaAnualPARp
from inewave.newave.modelos.dger import BlocoRestricoesFornecGas
from inewave.newave.modelos.dger import BlocoIncertezaGeracaoEolica
from inewave.newave.modelos.dger import BlocoIncertezaGeracaoSolar
from inewave.newave.modelos.dger import BlocoRepresentacaoIncerteza
from inewave.newave.modelos.dger import LeituraDGer


class DGer(Arquivo):
    """
    Classe para armazenar dados gerais de uma execução do NEWAVE.

    """
    def __init__(self, dados: DadosArquivo):
        super().__init__(dados)

    @classmethod
    def le_arquivo(cls,
                   diretorio: str,
                   nome_arquivo="dger.dat") -> 'DGer':
        """
        """
        leitor = LeituraDGer(diretorio)
        r = leitor.le_arquivo(nome_arquivo)
        return cls(r)

    def escreve_arquivo(self,
                        diretorio: str,
                        nome_arquivo="dger.dat"):
        """
        """
        escritor = Escrita(diretorio)
        escritor.escreve_arquivo(self._dados,
                                 nome_arquivo)

    def __le_por_tipo(self, tipo: Type[Bloco]) -> Any:
        for d in self._dados.blocos:
            if isinstance(d, tipo):
                return d.dados
        # Se não tem o dado, lança erro
        raise ValueError(f" Não foi encontrado o dado do tipo {tipo}")

    def __escreve_por_tipo(self,
                           tipo: Type[Bloco],
                           dado: Any):
        for d in self._dados.blocos:
            if isinstance(d, tipo):
                d.dados = dado
                return
        # Se não tem o dado, lança erro
        raise ValueError(f" Não foi encontrado o dado do tipo {tipo}")

    @property
    def nome_caso(self) -> str:
        """
        Configuração da linha número 1 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoNomeCaso)

    @nome_caso.setter
    def nome_caso(self, dado: str):
        self.__escreve_por_tipo(BlocoNomeCaso, dado)

    @property
    def tipo_execucao(self) -> int:
        """
        Configuração da linha número 2 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoTipoExecucao)

    @tipo_execucao.setter
    def tipo_execucao(self, dado: int):
        self.__escreve_por_tipo(BlocoTipoExecucao, dado)

    @property
    def duracao_periodo(self) -> int:
        """
        Configuração da linha número 3 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoDuracaoPeriodo)

    @duracao_periodo.setter
    def duracao_periodo(self, dado: int):
        self.__escreve_por_tipo(BlocoDuracaoPeriodo, dado)

    @property
    def num_anos_estudo(self) -> int:
        """
        Configuração da linha número 4 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoNumAnosEstudo)

    @num_anos_estudo.setter
    def num_anos_estudo(self, dado: int):
        self.__escreve_por_tipo(BlocoNumAnosEstudo, dado)

    @property
    def mes_inicio_pre_estudo(self) -> int:
        """
        Configuração da linha número 5 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoMesInicioPreEstudo)

    @mes_inicio_pre_estudo.setter
    def mes_inicio_pre_estudo(self, dado: int):
        self.__escreve_por_tipo(BlocoMesInicioPreEstudo, dado)

    @property
    def mes_inicio_estudo(self) -> int:
        """
        Configuração da linha número 6 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoMesInicioEstudo)

    @mes_inicio_estudo.setter
    def mes_inicio_estudo(self, dado: int):
        self.__escreve_por_tipo(BlocoMesInicioEstudo, dado)

    @property
    def ano_inicio_estudo(self) -> int:
        """
        Configuração da linha número 7 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoAnoInicioEstudo)

    @ano_inicio_estudo.setter
    def ano_inicio_estudo(self, dado: int):
        self.__escreve_por_tipo(BlocoAnoInicioEstudo, dado)

    @property
    def num_anos_pre_estudo(self) -> int:
        """
        Configuração da linha número 8 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoNumAnosPreEstudo)

    @num_anos_pre_estudo.setter
    def num_anos_pre_estudo(self, dado: int):
        self.__escreve_por_tipo(BlocoNumAnosPreEstudo, dado)

    @property
    def num_anos_pos_estudo(self) -> int:
        """
        Configuração da linha número 9 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoNumAnosPosEstudo)

    @num_anos_pos_estudo.setter
    def num_anos_pos_estudo(self, dado: int):
        self.__escreve_por_tipo(BlocoNumAnosPosEstudo, dado)

    @property
    def num_anos_pos_sim_final(self) -> int:
        """
        Configuração da linha número 10 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoNumAnosPosEstudoSimFinal)

    @num_anos_pos_sim_final.setter
    def num_anos_pos_sim_final(self, dado: int):
        self.__escreve_por_tipo(BlocoNumAnosPosEstudoSimFinal, dado)

    @property
    def imprime_dados(self) -> int:
        """
        Configuração da linha número 11 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoImprimeDados)

    @imprime_dados.setter
    def imprime_dados(self, dado: int):
        self.__escreve_por_tipo(BlocoImprimeDados, dado)

    @property
    def imprime_mercados(self) -> int:
        """
        Configuração da linha número 12 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoImprimeMercados)

    @imprime_mercados.setter
    def imprime_mercados(self, dado: int):
        self.__escreve_por_tipo(BlocoImprimeMercados, dado)

    @property
    def imprime_energias(self) -> int:
        """
        Configuração da linha número 13 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoImprimeEnergias)

    @imprime_energias.setter
    def imprime_energias(self, dado: int):
        self.__escreve_por_tipo(BlocoImprimeEnergias, dado)

    @property
    def imprime_modelo_estocastico(self) -> int:
        """
        Configuração da linha número 14 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoImprimeModeloEstocastico)

    @imprime_modelo_estocastico.setter
    def imprime_modelo_estocastico(self, dado: int):
        self.__escreve_por_tipo(BlocoImprimeModeloEstocastico, dado)

    @property
    def imprime_subsistema(self) -> int:
        """
        Configuração da linha número 15 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoImprimeSubsistema)

    @imprime_subsistema.setter
    def imprime_subsistema(self, dado: int):
        self.__escreve_por_tipo(BlocoImprimeSubsistema, dado)

    @property
    def num_max_iteracoes(self) -> int:
        """
        Configuração da linha número 16 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoNumMaxIteracoes)

    @num_max_iteracoes.setter
    def num_max_iteracoes(self, dado: int):
        self.__escreve_por_tipo(BlocoNumMaxIteracoes, dado)

    @property
    def num_forwards(self) -> int:
        """
        Configuração da linha número 17 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoNumForwards)

    @num_forwards.setter
    def num_forwards(self, dado: int):
        self.__escreve_por_tipo(BlocoNumForwards, dado)

    @property
    def num_aberturas(self) -> List[int]:
        """
        Configuração da linha número 18 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoNumAberturas)

    @num_aberturas.setter
    def num_aberturas(self, dado: List[int]):
        self.__escreve_por_tipo(BlocoNumAberturas, dado)

    @property
    def num_series_sinteticas(self) -> int:
        """
        Configuração da linha número 19 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoNumSeriesSinteticas)

    @num_series_sinteticas.setter
    def num_series_sinteticas(self, dado: int):
        self.__escreve_por_tipo(BlocoNumSeriesSinteticas, dado)

    @property
    def ordem_maxima_parp(self) -> int:
        """
        Configuração da linha número 20 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoOrdemMaximaPARp)

    @ordem_maxima_parp.setter
    def ordem_maxima_parp(self, dado: int):
        self.__escreve_por_tipo(BlocoOrdemMaximaPARp, dado)

    @property
    def ano_inicial_historico(self) -> List[int]:
        """
        Configuração da linha número 21 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoAnoInicialHistorico)

    @ano_inicial_historico.setter
    def ano_inicial_historico(self, dado: List[int]):
        self.__escreve_por_tipo(BlocoAnoInicialHistorico, dado)

    @property
    def calcula_volume_inicial(self) -> int:
        """
        Configuração da linha número 22 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoCalculaVolInicial)

    @calcula_volume_inicial.setter
    def calcula_volume_inicial(self, dado: int):
        self.__escreve_por_tipo(BlocoCalculaVolInicial, dado)

    @property
    def volume_inicial_por_subsistema(self) -> List[float]:
        """
        Configuração das linhas número 23 e 24 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoVolInicialSubsistema)

    @volume_inicial_por_subsistema.setter
    def volume_inicial_por_subsistema(self, dado: List[float]):
        self.__escreve_por_tipo(BlocoVolInicialSubsistema, dado)

    @property
    def tolerancia(self) -> float:
        """
        Configuração da linha número 25 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoTolerancia)

    @tolerancia.setter
    def tolerancia(self, dado: float):
        self.__escreve_por_tipo(BlocoTolerancia, dado)

    @property
    def taxa_de_desconto(self) -> float:
        """
        Configuração da linha número 26 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoTaxaDesconto)

    @taxa_de_desconto.setter
    def taxa_de_desconto(self, dado: float):
        self.__escreve_por_tipo(BlocoTaxaDesconto, dado)

    @property
    def tipo_simulacao_final(self) -> int:
        """
        Configuração da linha número 27 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoTipoSimFinal)

    @tipo_simulacao_final.setter
    def tipo_simulacao_final(self, dado: int):
        self.__escreve_por_tipo(BlocoTipoSimFinal, dado)

    @property
    def impressao_operacao(self) -> int:
        """
        Configuração da linha número 28 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoImpressaoOperacao)

    @impressao_operacao.setter
    def impressao_operacao(self, dado: int):
        self.__escreve_por_tipo(BlocoImpressaoOperacao, dado)

    @property
    def impressao_convergencia(self) -> int:
        """
        Configuração da linha número 29 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoImpressaoConvergencia)

    @impressao_convergencia.setter
    def impressao_convergencia(self, dado: int):
        self.__escreve_por_tipo(BlocoImpressaoConvergencia, dado)

    @property
    def intervalo_para_gravar(self) -> int:
        """
        Configuração da linha número 30 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoIntervaloGravar)

    @intervalo_para_gravar.setter
    def intervalo_para_gravar(self, dado: int):
        self.__escreve_por_tipo(BlocoIntervaloGravar, dado)

    @property
    def num_minimo_iteracoes(self) -> int:
        """
        Configuração da linha número 31 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoMinIteracoes)

    @num_minimo_iteracoes.setter
    def num_minimo_iteracoes(self, dado: int):
        self.__escreve_por_tipo(BlocoMinIteracoes, dado)

    @property
    def racionamento_preventivo(self) -> str:
        """
        Configuração da linha número 32 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoRacionamentoPreventivo)

    @racionamento_preventivo.setter
    def racionamento_preventivo(self, dado: str):
        self.__escreve_por_tipo(BlocoRacionamentoPreventivo, dado)

    @property
    def num_anos_manutencao_utes(self) -> int:
        """
        Configuração da linha número 33 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoNumAnosManutUTE)

    @num_anos_manutencao_utes.setter
    def num_anos_manutencao_utes(self, dado: int):
        self.__escreve_por_tipo(BlocoNumAnosManutUTE, dado)

    @property
    def tendencia_hidrologica(self) -> List[int]:
        """
        Configuração da linha número 34 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoTendenciaHidrologica)

    @tendencia_hidrologica.setter
    def tendencia_hidrologica(self, dado: List[int]):
        self.__escreve_por_tipo(BlocoTendenciaHidrologica, dado)

    @property
    def restricao_itaipu(self) -> int:
        """
        Configuração da linha número 35 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoRestricaoItaipu)

    @restricao_itaipu.setter
    def restricao_itaipu(self, dado: int):
        self.__escreve_por_tipo(BlocoRestricaoItaipu, dado)

    @property
    def bid(self) -> int:
        """
        Configuração da linha número 36 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoBid)

    @bid.setter
    def bid(self, dado: int):
        self.__escreve_por_tipo(BlocoBid, dado)

    @property
    def perdas_rede_transmissao(self) -> int:
        """
        Configuração da linha número 37 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoPerdasTransmissao)

    @perdas_rede_transmissao.setter
    def perdas_rede_transmissao(self, dado: int):
        self.__escreve_por_tipo(BlocoPerdasTransmissao, dado)

    @property
    def el_nino(self) -> int:
        """
        Configuração da linha número 38 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoElNino)

    @el_nino.setter
    def el_nino(self, dado: int):
        self.__escreve_por_tipo(BlocoElNino, dado)

    @property
    def enso(self) -> int:
        """
        Configuração da linha número 39 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoEnso)

    @enso.setter
    def enso(self, dado: int):
        self.__escreve_por_tipo(BlocoEnso, dado)

    @property
    def duracao_por_patamar(self) -> int:
        """
        Configuração da linha número 40 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoDuracaoPorPatamar)

    @duracao_por_patamar.setter
    def duracao_por_patamar(self, dado: int):
        self.__escreve_por_tipo(BlocoDuracaoPorPatamar, dado)

    @property
    def outros_usos_da_agua(self) -> int:
        """
        Configuração da linha número 41 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoOutrosUsosAgua)

    @outros_usos_da_agua.setter
    def outros_usos_da_agua(self, dado: int):
        self.__escreve_por_tipo(BlocoOutrosUsosAgua, dado)

    @property
    def correcao_desvio(self) -> int:
        """
        Configuração da linha número 42 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoCorrecaoDesvio)

    @correcao_desvio.setter
    def correcao_desvio(self, dado: int):
        self.__escreve_por_tipo(BlocoCorrecaoDesvio, dado)

    @property
    def curva_aversao(self) -> int:
        """
        Configuração da linha número 43 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoCurvaAversao)

    @curva_aversao.setter
    def curva_aversao(self, dado: int):
        self.__escreve_por_tipo(BlocoCurvaAversao, dado)

    @property
    def tipo_geracao_enas(self) -> int:
        """
        Configuração da linha número 44 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoTipoGeracaoENA)

    @tipo_geracao_enas.setter
    def tipo_geracao_enas(self, dado: int):
        self.__escreve_por_tipo(BlocoTipoGeracaoENA, dado)

    @property
    def risco_deficit(self) -> List[float]:
        """
        Configuração da linha número 45 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoRiscoDeficit)

    @risco_deficit.setter
    def risco_deficit(self, dado: List[float]):
        self.__escreve_por_tipo(BlocoRiscoDeficit, dado)

    @property
    def iteracao_para_simulacao_final(self) -> int:
        """
        Configuração da linha número 46 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoIteracaoParaSimFinal)

    @iteracao_para_simulacao_final.setter
    def iteracao_para_simulacao_final(self, dado: int):
        self.__escreve_por_tipo(BlocoIteracaoParaSimFinal, dado)

    @property
    def agrupamento_livre(self) -> int:
        """
        Configuração da linha número 47 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoAgrupamentoLivre)

    @agrupamento_livre.setter
    def agrupamento_livre(self, dado: int):
        self.__escreve_por_tipo(BlocoAgrupamentoLivre, dado)

    @property
    def equalizacao_penal_itercambio(self) -> int:
        """
        Configuração da linha número 48 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoEqualizacaoPenalInt)

    @equalizacao_penal_itercambio.setter
    def equalizacao_penal_itercambio(self, dado: int):
        self.__escreve_por_tipo(BlocoEqualizacaoPenalInt, dado)

    @property
    def representacao_submotorizacao(self) -> int:
        """
        Configuração da linha número 49 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoRepresentacaoSubmot)

    @representacao_submotorizacao.setter
    def representacao_submotorizacao(self, dado: int):
        self.__escreve_por_tipo(BlocoRepresentacaoSubmot, dado)

    @property
    def ordenacao_automatica(self) -> int:
        """
        Configuração da linha número 50 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoOrdenacaoAutomatica)

    @ordenacao_automatica.setter
    def ordenacao_automatica(self, dado: int):
        self.__escreve_por_tipo(BlocoOrdenacaoAutomatica, dado)

    @property
    def considera_carga_adicional(self) -> int:
        """
        Configuração da linha número 51 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoConsideraCargaAdicional)

    @considera_carga_adicional.setter
    def considera_carga_adicional(self, dado: int):
        self.__escreve_por_tipo(BlocoConsideraCargaAdicional, dado)

    @property
    def delta_zsup(self) -> float:
        """
        Configuração da linha número 52 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoDeltaZSUP)

    @delta_zsup.setter
    def delta_zsup(self, dado: float):
        self.__escreve_por_tipo(BlocoDeltaZSUP, dado)

    @property
    def delta_zinf(self) -> float:
        """
        Configuração da linha número 53 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoDeltaZINF)

    @delta_zinf.setter
    def delta_zinf(self, dado: float):
        self.__escreve_por_tipo(BlocoDeltaZINF, dado)

    @property
    def deltas_consecutivos(self) -> int:
        """
        Configuração da linha número 54 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoDeltasConsecutivos)

    @deltas_consecutivos.setter
    def deltas_consecutivos(self, dado: int):
        self.__escreve_por_tipo(BlocoDeltasConsecutivos, dado)

    @property
    def despacho_antecipado_gnl(self) -> int:
        """
        Configuração da linha número 55 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoDespachoAntecipadoGNL)

    @despacho_antecipado_gnl.setter
    def despacho_antecipado_gnl(self, dado: int):
        self.__escreve_por_tipo(BlocoDespachoAntecipadoGNL, dado)

    @property
    def modif_automatica_adterm(self) -> int:
        """
        Configuração da linha número 56 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoModifAutomaticaAdTerm)

    @modif_automatica_adterm.setter
    def modif_automatica_adterm(self, dado: int):
        self.__escreve_por_tipo(BlocoModifAutomaticaAdTerm, dado)

    @property
    def considera_ghmin(self) -> int:
        """
        Configuração da linha número 57 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoGeracaoHidraulicaMin)

    @considera_ghmin.setter
    def considera_ghmin(self, dado: int):
        self.__escreve_por_tipo(BlocoGeracaoHidraulicaMin, dado)

    @property
    def simulacao_final_com_data(self) -> int:
        """
        Configuração da linha número 58 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoSimFinalComData)

    @simulacao_final_com_data.setter
    def simulacao_final_com_data(self, dado: int):
        self.__escreve_por_tipo(BlocoSimFinalComData, dado)

    @property
    def gerenciamento_pls(self) -> List[int]:
        """
        Configuração da linha número 59 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoGerenciamentoPLs)

    @gerenciamento_pls.setter
    def gerenciamento_pls(self, dado: List[int]):
        self.__escreve_por_tipo(BlocoGerenciamentoPLs, dado)

    @property
    def sar(self) -> int:
        """
        Configuração da linha número 60 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoSAR)

    @sar.setter
    def sar(self, dado: int):
        self.__escreve_por_tipo(BlocoSAR, dado)

    @property
    def cvar(self) -> int:
        """
        Configuração da linha número 61 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoCVAR)

    @cvar.setter
    def cvar(self, dado: int):
        self.__escreve_por_tipo(BlocoCVAR, dado)

    @property
    def considera_zsup_min_convergencia(self) -> int:
        """
        Configuração da linha número 62 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoZSUPMinConvergencia)

    @considera_zsup_min_convergencia.setter
    def considera_zsup_min_convergencia(self, dado: int):
        self.__escreve_por_tipo(BlocoZSUPMinConvergencia, dado)

    @property
    def desconsidera_vazao_minima(self) -> int:
        """
        Configuração da linha número 63 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoDesconsideraVazaoMinima)

    @desconsidera_vazao_minima.setter
    def desconsidera_vazao_minima(self, dado: int):
        self.__escreve_por_tipo(BlocoDesconsideraVazaoMinima, dado)

    @property
    def restricoes_eletricas(self) -> int:
        """
        Configuração da linha número 64 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoRestricoesEletricas)

    @restricoes_eletricas.setter
    def restricoes_eletricas(self, dado: int):
        self.__escreve_por_tipo(BlocoRestricoesEletricas, dado)

    @property
    def selecao_de_cortes(self) -> int:
        """
        Configuração da linha número 65 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoSelecaoCortes)

    @selecao_de_cortes.setter
    def selecao_de_cortes(self, dado: int):
        self.__escreve_por_tipo(BlocoSelecaoCortes, dado)

    @property
    def janela_de_cortes(self) -> int:
        """
        Configuração da linha número 66 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoJanelaCortes)

    @janela_de_cortes.setter
    def janela_de_cortes(self, dado: int):
        self.__escreve_por_tipo(BlocoJanelaCortes, dado)

    @property
    def reamostragem_cenarios(self) -> List[int]:
        """
        Configuração da linha número 67 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoReamostragemCenarios)

    @reamostragem_cenarios.setter
    def reamostragem_cenarios(self, dado: List[int]):
        self.__escreve_por_tipo(BlocoReamostragemCenarios, dado)

    @property
    def converge_no_zero(self) -> int:
        """
        Configuração da linha número 68 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoConvergeNoZero)

    @converge_no_zero.setter
    def converge_no_zero(self, dado: int):
        self.__escreve_por_tipo(BlocoConvergeNoZero, dado)

    @property
    def consulta_fcf(self) -> int:
        """
        Configuração da linha número 69 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoConsultaFCF)

    @consulta_fcf.setter
    def consulta_fcf(self, dado: int):
        self.__escreve_por_tipo(BlocoConsultaFCF, dado)

    @property
    def impressao_ena(self) -> int:
        """
        Configuração da linha número 70 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoImpressaoENA)

    @impressao_ena.setter
    def impressao_ena(self, dado: int):
        self.__escreve_por_tipo(BlocoImpressaoENA, dado)

    @property
    def impressao_cortes_ativos_sim_final(self) -> str:
        """
        Configuração da linha número 71 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoImpressaoCortesAtivosSimFinal)

    @impressao_cortes_ativos_sim_final.setter
    def impressao_cortes_ativos_sim_final(self, dado: str):
        self.__escreve_por_tipo(BlocoImpressaoCortesAtivosSimFinal,
                                dado)

    @property
    def representacao_agregacao(self) -> int:
        """
        Configuração da linha número 72 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoRepresentacaoAgregacao)

    @representacao_agregacao.setter
    def representacao_agregacao(self, dado: int):
        self.__escreve_por_tipo(BlocoRepresentacaoAgregacao, dado)

    @property
    def matriz_correlacao_espacial(self) -> int:
        """
        Configuração da linha número 73 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoMatrizCorrelacaoEspacial)

    @matriz_correlacao_espacial.setter
    def matriz_correlacao_espacial(self, dado: int):
        self.__escreve_por_tipo(BlocoMatrizCorrelacaoEspacial, dado)

    @property
    def desconsidera_convergencia_estatistica(self) -> int:
        """
        Configuração da linha número 74 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoDesconsideraConvEstatistica)

    @desconsidera_convergencia_estatistica.setter
    def desconsidera_convergencia_estatistica(self, dado: int):
        self.__escreve_por_tipo(BlocoDesconsideraConvEstatistica,
                                dado)

    @property
    def momento_reamostragem(self) -> int:
        """
        Configuração da linha número 75 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoMomentoReamostragem)

    @momento_reamostragem.setter
    def momento_reamostragem(self, dado: int):
        self.__escreve_por_tipo(BlocoMomentoReamostragem, dado)

    @property
    def mantem_arquivos_energias(self) -> int:
        """
        Configuração da linha número 76 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoMantemArquivosEnergias)

    @mantem_arquivos_energias.setter
    def mantem_arquivos_energias(self, dado: int):
        self.__escreve_por_tipo(BlocoMantemArquivosEnergias, dado)

    @property
    def inicio_teste_convergencia(self) -> int:
        """
        Configuração da linha número 77 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoInicioTesteConvergencia)

    @inicio_teste_convergencia.setter
    def inicio_teste_convergencia(self, dado: int):
        self.__escreve_por_tipo(BlocoInicioTesteConvergencia, dado)

    @property
    def sazonaliza_vmint(self) -> int:
        """
        Configuração da linha número 78 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoSazonalizarVminT)

    @sazonaliza_vmint.setter
    def sazonaliza_vmint(self, dado: int):
        self.__escreve_por_tipo(BlocoSazonalizarVminT, dado)

    @property
    def sazonaliza_vmaxt(self) -> int:
        """
        Configuração da linha número 79 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoSazonalizarVmaxT)

    @sazonaliza_vmaxt.setter
    def sazonaliza_vmaxt(self, dado: int):
        self.__escreve_por_tipo(BlocoSazonalizarVmaxT, dado)

    @property
    def sazonaliza_vminp(self) -> int:
        """
        Configuração da linha número 80 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoSazonalizarVminP)

    @sazonaliza_vminp.setter
    def sazonaliza_vminp(self, dado: int):
        self.__escreve_por_tipo(BlocoSazonalizarVminP, dado)

    @property
    def sazonaliza_cfuga_cmont(self) -> int:
        """
        Configuração da linha número 81 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoSazonalizarCfugaCmont)

    @sazonaliza_cfuga_cmont.setter
    def sazonaliza_cfuga_cmont(self, dado: int):
        self.__escreve_por_tipo(BlocoSazonalizarCfugaCmont, dado)

    @property
    def restricoes_emissao_gee(self) -> int:
        """
        Configuração da linha número 82 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoRestricoesEmissaoGEE)

    @restricoes_emissao_gee.setter
    def restricoes_emissao_gee(self, dado: int):
        self.__escreve_por_tipo(BlocoRestricoesEmissaoGEE, dado)

    @property
    def afluencia_anual_parp(self) -> int:
        """
        Configuração da linha número 83 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoAfluenciaAnualPARp)

    @afluencia_anual_parp.setter
    def afluencia_anual_parp(self, dado: int):
        self.__escreve_por_tipo(BlocoAfluenciaAnualPARp, dado)

    @property
    def restricoes_fornecimento_gas(self) -> int:
        """
        Configuração da linha número 84 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoRestricoesFornecGas)

    @restricoes_fornecimento_gas.setter
    def restricoes_fornecimento_gas(self, dado: int):
        self.__escreve_por_tipo(BlocoRestricoesFornecGas, dado)

    @property
    def incerteza_geracao_eolica(self) -> int:
        """
        Configuração da linha número 85 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoIncertezaGeracaoEolica)

    @incerteza_geracao_eolica.setter
    def incerteza_geracao_eolica(self, dado: int):
        self.__escreve_por_tipo(BlocoIncertezaGeracaoEolica, dado)

    @property
    def incerteza_geracao_solar(self) -> int:
        """
        Configuração da linha número 86 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoIncertezaGeracaoSolar)

    @incerteza_geracao_solar.setter
    def incerteza_geracao_solar(self, dado: int):
        self.__escreve_por_tipo(BlocoIncertezaGeracaoSolar, dado)

    @property
    def representacao_incertezas(self) -> int:
        """
        Configuração da linha número 87 do arquivo `dger.dat`.
        """
        return self.__le_por_tipo(BlocoRepresentacaoIncerteza)

    @representacao_incertezas.setter
    def representacao_incertezas(self, dado: int):
        self.__escreve_por_tipo(BlocoRepresentacaoIncerteza, dado)



# Rotinas de testes associadas ao arquivo eafbm00x.out do NWLISTOP
from inewave.nwlistop.eafbm00 import EafbM00


sub_teste = "SUDESTE"
eaf = EafbM00.le_arquivo("tests/_arquivos", "eafbm00test.out")


def test_leitura():
    assert sub_teste == eaf.submercado


def test_eq_eafbm00():
    eaf2 = EafbM00.le_arquivo("tests/_arquivos", "eafbm00test.out")
    assert eaf == eaf2


def test_neq_eafbm00():
    eaf2 = EafbM00.le_arquivo("tests/_arquivos", "eafbm00test.out")
    eaf2.energias.iloc[0, 0] = -1
    assert eaf != eaf2



# Imports do próprio módulo
from inewave._utils.leitura import LeituraCSV
# Imports de módulos externos


class LeituraMediasSIN(LeituraCSV):
    """
    Realiza a leitura do arquivo MEDIAS-SIN.CSV
    existente em um diretório de saídas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos do arquivo MEDIAS-SIN.CSV, construindo um
    objeto `MediasSIN` cujas informações são as mesmas do arquivo.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.

    """
    def __init__(self,
                 diretorio: str) -> None:
        super().__init__(diretorio)

    def processa_dados_lidos(self):
        # Renomeia as colunas
        cols_atuais = list(self._dados.columns)
        cols_novas = [c.strip() for c in cols_atuais]
        self._dados.columns = cols_novas
        # Exclui a primeira e a última coluna (todos 0 | em branco)
        self._dados.drop(columns=[cols_novas[0], cols_novas[-1]],
                         inplace=True)



import numpy as np  # type: ignore
from inewave._utils.dadosarquivo import DadosArquivo
from inewave._utils.arquivo import Arquivo
from inewave.newave.modelos.dsvagua import BlocoDsvUHE, LeituraDSVAgua
from inewave._utils.escrita import Escrita


class DSVAgua(Arquivo):
    """
    Armazena os dados de entrada do NEWAVE referentes aos
    desvios de água por usina.

    Esta classe lida com informações de entrada fornecidas ao NEWAVE e
    que podem ser modificadas através do arquivo `dsvagua.dat`.

    """

    def __init__(self,
                 dados: DadosArquivo):
        super().__init__(dados)
        # Interpreta o resultado da leitura
        val = True
        msg = "Erro na criação de DSVAgua: "
        if len(dados.blocos) == 1:
            bloco = dados.blocos[0]
            if isinstance(bloco, BlocoDsvUHE):
                self.__bloco = bloco
            else:
                msg += (f"O bloco deve ser do tipo {BlocoDsvUHE}, " +
                        f"mas foi fornecido do tipo {type(bloco)}")
                val = False
        else:
            msg += "Deve ser fornecido exatamente 1 bloco para DSVAgua"
            val = False
        if not val:
            raise TypeError(msg)

    @classmethod
    def le_arquivo(cls,
                   diretorio: str,
                   nome_arquivo="dsvagua.dat") -> 'DSVAgua':
        leitor = LeituraDSVAgua(diretorio)
        r = leitor.le_arquivo(nome_arquivo)
        return cls(r)

    def escreve_arquivo(self,
                        diretorio: str,
                        nome_arquivo="dsvagua.dat"):
        escritor = Escrita(diretorio)
        escritor.escreve_arquivo(self._dados, nome_arquivo)

    @property
    def desvios(self) -> np.ndarray:
        return self.__bloco.dados



# Rotinas de testes associadas ao arquivo pmo.dat do NEWAVE
from inewave.newave.pmo import PMO
from inewave.config import REES
import numpy as np  # type: ignore


pmo = PMO.le_arquivo("tests/_arquivos")


def test_eq_pmo():
    pmo2 = PMO.le_arquivo("tests/_arquivos")
    assert pmo == pmo2


def test_neq_pmo():
    pmo2 = PMO.le_arquivo("tests/_arquivos")
    entrada = pmo2.configuracoes_qualquer_modificacao
    pmo2.configuracoes_alteracao_potencia = entrada
    assert pmo != pmo2


def test_leitura_configs():
    configs = pmo.configuracoes_alteracao_potencia
    assert configs.shape[0] == 10
    for ano in range(configs.shape[0]):
        assert np.all(configs[ano, :] > 0)


def test_leitura_retas_perdas():
    perdas = pmo.retas_perdas_engolimento
    assert len(perdas.keys()) == len(REES)


def test_leitura_convergencia():
    convergencia = pmo.convergencia
    assert len(list(convergencia.index)) == 135


def test_leitura_ens():
    risco_ens = pmo.risco_deficit_ens
    assert len(list(risco_ens.index)) == 5
    assert risco_ens.iloc[4, 1] == 0.05
    assert risco_ens.iloc[4, 2] == 0.4
    assert risco_ens.iloc[4, 3] == 0.05
    assert risco_ens.iloc[4, 4] == 0.3


def test_leitura_tabelas_custos():
    custo_series = pmo.custo_operacao_series_simuladas
    valor_esp = pmo.valor_esperado_periodo_estudo
    custo_ref = pmo.custo_operacao_referenciado_primeiro_mes
    assert custo_series.iloc[0, 0] == 18951.33
    assert valor_esp.iloc[0, 0] == 18951.33
    assert custo_ref.iloc[0, 0] == 18951.33


# # def test_eco_dger_pmo():
# #     leitor2 = LeituraDGer("tests/_arquivos")
# #     leitor2.le_arquivo()
# #     leitor.pmo.dados_gerais.eq_eco_saida(leitor2.dger)



# Rotinas de testes associadas ao arquivo dsvagua.dat do NEWAVE
from inewave.newave.dsvagua import DSVAgua


dsv = DSVAgua.le_arquivo("tests/_arquivos")


def test_leitura():
    assert dsv.desvios.shape[0] > 0


def test_escrita_e_leitura():
    dsv.escreve_arquivo("tests/_saidas")
    dsv2 = DSVAgua.le_arquivo("tests/_saidas")
    assert dsv == dsv2


def test_eq_dsvagua():
    dsv2 = DSVAgua.le_arquivo("tests/_arquivos")
    assert dsv == dsv2


# def test_neq_dsvagua():
#     dsv2 = DSVAgua.le_arquivo("tests/_arquivos")
#     dsv2.le_arquivo()
#     dsv2.dsvagua.tabela[0, 0] = 1e-6
#     assert dsv2.dsvagua != leitor.dsvagua
#     assert dsv2.dsvagua is not None



from inewave._utils.arquivo import Arquivo
from inewave._utils.dadosarquivo import DadosArquivo
from inewave._utils.escrita import Escrita
from inewave.newave.modelos.modif import LeituraModif


class Modif(Arquivo):
    """
    Armazena os dados de entrada do NEWAVE referentes às alterações nas
    configurações das usinas hidroelétricas.

    **Parâmetros**

    """

    def __init__(self,
                 dados: DadosArquivo) -> None:
        super().__init__(dados)

    # Override
    @classmethod
    def le_arquivo(cls,
                   diretorio: str,
                   nome_arquivo="modif.dat") -> 'Modif':
        """
        """
        leitor = LeituraModif(diretorio)
        r = leitor.le_arquivo(nome_arquivo)
        return cls(r)

    def escreve_arquivo(self,
                        diretorio: str,
                        nome_arquivo="modif.dat"):
        """
        """
        escritor = Escrita(diretorio)
        escritor.escreve_arquivo(self._dados,
                                 nome_arquivo)



from inewave._utils.arquivo import Arquivo
from inewave._utils.dadosarquivo import DadosArquivo
from inewave.nwlistop.modelos.earmfp00 import LeituraEarmfp00

import pandas as pd  # type: ignore


class Earmfp00(Arquivo):
    """
    Armazena os dados das saídas referentes às armazenadas finais,
    por REE em valores absolutos.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `earmfp00x.out`, onde x varia conforme o
    REE em questão.

    """
    def __init__(self,
                 dados: DadosArquivo):
        super().__init__(dados)

    # Override
    @classmethod
    def le_arquivo(cls,
                   diretorio: str,
                   nome_arquivo="earmfp001.out") -> 'Earmfp00':
        """
        """
        leitor = LeituraEarmfp00(diretorio)
        r = leitor.le_arquivo(nome_arquivo)
        return cls(r)

    @property
    def ree(self) -> str:
        """
        Tabela com o REE associado ao arquivo lido.

         **Retorna**

        `str`
        """
        return self._blocos[0].dados[0]

    @property
    def energias(self) -> pd.DataFrame:
        """
        Tabela com as energias armazenadas finais por série e
        por mês/ano de estudo.

         **Retorna**

        `pd.DataFrame`
        """
        return self._blocos[0].dados[1]



from inewave.nwlistcf.nwlistcf import Nwlistcf

fcf = Nwlistcf.le_arquivo("./tests/_arquivos")


def test_leitura():
    assert len(fcf.registros) > 0


def test_eq_nwlistcf():
    fcf2 = Nwlistcf.le_arquivo("./tests/_arquivos")
    assert fcf2 == fcf


# def test_neq_nwlistcf():
#     leitor2 = LeituraNwlistcf("./tests/_arquivos")
#     leitor2.le_arquivo()
#     leitor2.nwlistcf.registros[5] = {}
#     assert leitor2.nwlistcf != leitor.nwlistcf



# Inclui os membros

from .cmarg00 import Cmarg00  # noqa
from .eafbm00 import EafbM00  # noqa
from .earmfp00 import Earmfp00  # noqa
from .earmfpm00 import EarmfpM00  # noqa
from .ghtot00 import GHTot00  # noqa
from .ghtotm00 import GHTotM00  # noqa
from .gttotsin import GTTotSIN  # noqa
from .mediasmerc import MediasMerc  # noqa
from .mediassin import MediasSIN  # noqa
from .vagua00 import VAgua00  # noqa



from inewave.config import REES
from inewave._utils.arquivo import Arquivo
from inewave._utils.bloco import Bloco
from inewave._utils.dadosarquivo import DadosArquivo
from inewave.newave.modelos.pmo import BlocoEcoDgerPMO
from inewave.newave.modelos.pmo import BlocoConvergenciaPMO
from inewave.newave.modelos.pmo import BlocoConfiguracoesExpansaoPMO
from inewave.newave.modelos.pmo import BlocoMARSPMO
from inewave.newave.modelos.pmo import BlocoRiscoDeficitENSPMO
from inewave.newave.modelos.pmo import BlocoCustoOperacaoPMO
from inewave.newave.modelos.pmo import LeituraPMO

from typing import Dict, List, Type
import numpy as np  # type: ignore
import pandas as pd  # type: ignore


class EnergiasAfluentesPMO:
    """
    Armazena as informações de energias
    afluentes anteriores ao estudo contidas no arquivo `pmo.dat`.

    **Parâmetros**
    """
    def __init__(self):
        # TODO - estruturar as propriedades
        pass

    def __eq__(self, o: object):
        return True


class EnergiaFioLiquidaREEPMO:
    """
    Armazena as informações de energias a fio d'água líquidas
    para cada REE existentes no arquivo `pmo.dat`,
    quando feita a simulação completa.

    **Parâmetros**

    - tabela: `np.ndarray`

    """

    def __init__(self,
                 tabela: np.ndarray):
        self.tabela = tabela

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre EnergiaFioLiquidaREEPMO avalia todos os campos.
        """
        if not isinstance(o, EnergiaFioLiquidaREEPMO):
            return False
        e: EnergiaFioLiquidaREEPMO = o
        return np.array_equal(self.tabela, e.tabela)


class ConfiguracoesExpansaoPMO:
    """
    Armazena as informações das configurações válidas para cada
    mês do estudo contidas no arquivo `pmo.dat`.

    **Parâmetros**

    - tabela: `np.ndarray`
    """
    def __init__(self,
                 tabela: np.ndarray):
        self.tabela = tabela

    def __eq__(self, o: object):
        """
        A igualdade entre ConfiguracoesExpansaoPMO avalia todos os campos.
        """
        if not isinstance(o, ConfiguracoesExpansaoPMO):
            return False
        e: ConfiguracoesExpansaoPMO = o
        return np.array_equal(self.tabela, e.tabela)

    @property
    def configs_por_ano(self) -> Dict[int, np.ndarray]:
        """
        Configurações ativas para serem consideradas em cada
        mês de estudo, organizada por ano.

        **Retorna**
        `Dict[int, np.ndarray]`
        """
        dict_configs: Dict[int, np.ndarray] = {}
        for lin in range(self.tabela.shape[0]):
            dict_configs[self.tabela[lin, 0]] = self.tabela[lin, 1:]

        return dict_configs


class RetasPerdasEngolimentoREEPMO:
    """
    Armazena as retas que modelam as perdas por engolimento
    máximo para cada REE, existentes no arquivo `pmo.dat`, quando
    feita a simulação completa.

    **Parâmetros**

    - tabela: `np.ndarray`

    """
    def __init__(self,
                 tabela: np.ndarray):
        self.tabela = tabela

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre RetasPerdasEngolimentoREEPMO
        avalia todos os campos.
        """
        if not isinstance(o, RetasPerdasEngolimentoREEPMO):
            return False
        e: RetasPerdasEngolimentoREEPMO = o
        return np.array_equal(self.tabela, e.tabela)

    def funcao_perdas(self,
                      ree: int,
                      energia_bruta: float) -> float:
        """
        Valor da função de perdas composta pelas retas
        para um determinado valor de energia bruta fornecido.

        **Parâmetros**

        - ree: `int`
        - energia_bruta: `float`

        **Retorna**
        `float`

        """
        n_lin = self.tabela.shape[0]
        # Calcula o valor das três retas na ordenada dada e
        # retorna o maior deles.
        energia = energia_bruta * np.ones((n_lin, 1))
        perdas: np.ndarray = np.multiply(self.tabela[ree-1, :, 1],
                                         energia) + self.tabela[ree-1, :, 2]
        return float(np.max(perdas))


class DemandaLiquidaEnergiaPMO:
    """
    Armazena as informações de demandas
    líquidas de energia contidas no arquivo `pmo.dat`.

    **Parâmetros**
    """
    def __init__(self):
        # TODO - estruturar as propriedades
        pass


class ConvergenciaPMO:
    """
    Armazena as informações do relatório de convergência do NEWAVE
    existentes no arquivo `pmo.dat`.

    **Parâmetros**

    - anos_estudo: `List[int]`
    - tabela: `np.ndarray`
    """
    def __init__(self,
                 tabela: np.ndarray):
        self.tabela = tabela

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre ConvergenciaPMO
        avalia todos os campos.
        """
        if not isinstance(o, ConvergenciaPMO):
            return False
        e: ConvergenciaPMO = o
        return np.array_equal(self.tabela, e.tabela)

    @property
    def zinf(self) -> Dict[int, List[float]]:
        """
        Limite inferior da função objetivo na PDDE (Zinf).

        **Retorna**

        `Dict[int, List[float]]`

        **Sobre**

        O número de chaves do dicionário é igual ao de iterações.
        """
        zinfs: Dict[int, List[float]] = {}
        for lin in range(self.tabela.shape[0]):
            it = int(self.tabela[lin, 0])
            if it not in zinfs:
                zinfs[it] = []
            zinfs[it].append(self.tabela[lin, 2])
        return zinfs

    @property
    def zsup(self) -> Dict[int, List[float]]:
        """
        Limite superior da função objetivo na PDDE (Zsup).

        **Retorna**

        `Dict[int, List[float]]`

        **Sobre**

        O número de chaves do dicionário é igual ao de iterações.
        """
        zsups: Dict[int, List[float]] = {}
        for lin in range(self.tabela.shape[0]):
            it = int(self.tabela[lin, 0])
            if it not in zsups:
                zsups[it] = []
            zsups[it].append(self.tabela[lin, 4])
        return zsups

    @property
    def tempos_execucao(self) -> List[float]:
        """
        Tempo de execução de cada iteração em segundos.

        **Retorna**

        `List[float]`

        **Sobre**

        O número de elementos da lista é igual ao de iterações.
        """
        tempos: List[float] = []
        ultima_it = -1
        for lin in range(self.tabela.shape[0]):
            it = self.tabela[lin, 0]
            if it == ultima_it:
                continue
            tempos.append(self.tabela[lin, -1])
            ultima_it = it
        return tempos


class RiscoDeficitENSPMO:
    """
    Armazena as informações risco de déficit e valores esperados
    de energia não supridacontidas no arquivo `pmo.dat`.

    **Parâmetros**

    - anos_estudo: `List[int]`
    - tabela: `np.ndarray`
    """
    def __init__(self,
                 anos_estudo: List[int],
                 tabela: np.ndarray):
        # São impressos no pmo.dat os subsistemas em
        # uma ordem pré-determinada
        self.subsistemas = ["SUDESTE",
                            "SUL",
                            "NORDESTE",
                            "NORTE"]
        self.anos_estudo = anos_estudo
        self.tabela = tabela

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre RiscoDeficitENSPMO
        avalia todos os campos.
        """
        if not isinstance(o, RiscoDeficitENSPMO):
            return False
        e: RiscoDeficitENSPMO = o
        eq_anos = all([a == b
                       for (a, b) in zip(self.anos_estudo,
                                         e.anos_estudo)])
        eq_tabela = np.array_equal(self.tabela, e.tabela)
        return eq_anos and eq_tabela

    @property
    def riscos_por_subsistema_e_ano(self) -> Dict[str,
                                                  Dict[int,
                                                       float]]:
        """
        Riscos de déficit agrupados por
        subsistema e ano de estudo.

        **Retorna**

        `Dict[str, Dict[int, float]]`

        **Sobre**

        O acesso é feito com
        [sub][ano] e o valor fornecido é em percentual, de 0 a 100.
        """
        riscos: Dict[str, Dict[int, float]] = {}
        for j, sub in enumerate(self.subsistemas):
            if sub not in riscos:
                riscos[sub] = {}
            for i, ano in enumerate(self.anos_estudo):
                riscos[sub][ano] = self.tabela[i, 2*j]
        return riscos

    @property
    def ens_por_subsistema_e_ano(self) -> Dict[str,
                                               Dict[int,
                                                    float]]:
        """
        Energias não supridas agrupadas por
        subsistema e ano de estudo.

        **Retorna**

        `Dict[str, Dict[int, float]]`

        **Sobre**

        O acesso é feito com
        [sub][ano] e o valor fornecido é em MWmes.
        """
        energias: Dict[str, Dict[int, float]] = {}
        for j, sub in enumerate(self.subsistemas):
            if sub not in energias:
                energias[sub] = {}
            for i, ano in enumerate(self.anos_estudo):
                energias[sub][ano] = self.tabela[i, 2*j + 1]
        return energias


class CustoOperacaoPMO:
    """
    Armazena as informações do relatório
    de custo de operação, disponível no arquivo `pmo.dat`.

    Esta classe armazena uma das tabelas existentes ao final do arquivo
    `pmo.dat`, contendo os custos, os desvios-padrão e a participação
    em percentual de cada componente de custo.

    A tabela de custos é armazenada através de uma array
    em `NumPy`, para otimizar cálculos futuros e espaço ocupado
    em memória. A tabela interna é transformada em dicionários
    e outras estruturas de dados mais palpáveis através das propriedades
    da própria classe.

    **Parâmetros**

    - custos: `np.ndarray`

    """
    def __init__(self,
                 custos: np.ndarray):
        self.custos = custos
        # TODO - fazer as @property de cada uma individualmente

        # self.geracao_termica = geracao_termica
        # self.deficit = deficit
        # self.vertimento = vertimento
        # self.excesso_energia = excesso_energia
        # self.violacao_car = violacao_car
        # self.violacao_sar = violacao_sar
        # self.violacao_outro_usos = violacao_outro_usos
        # self.violacao_evmin = violacao_evmin
        # self.violacao_vzmin = violacao_vzmin
        # self.intercambio = intercambio
        # self.violacao_intercambio_min = violacao_intercambio_min
        # self.vertimento_fio_nao_turbin = vertimento_fio_nao_turbin
        # self.violacao_ghmin = violacao_ghmin
        # self.violacao_ghmin_usina = violacao_ghmin_usina
        # self.violacao_retirada = violacao_retirada
        # self.violacao_emissao_gee = violacao_emissao_gee

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre CustoOperacaoPMO
        avalia todos os campos.
        """
        if not isinstance(o, CustoOperacaoPMO):
            return False
        e: CustoOperacaoPMO = o
        return np.array_equal(self.custos, e.custos)

    @property
    def geracao_termica(self) -> float:
        """
        Parcela do custo de operação devido à geração térmica,
        em MMR$.

        **Retorna**

        `float`
        """
        return float(self.custos[0, 0])

    @property
    def deficit(self) -> float:
        """
        Parcela do custo de operação devido ao déficit, em MMR$.

        **Retorna**

        `float`
        """
        return float(self.custos[1, 0])

    @property
    def custos_seguranca(self) -> float:
        """
        Parcela do custo de operação devido a violações de segurança,
        CAR e SAR, em MMR$.

        **Retorna**

        `float`
        """
        return np.sum(self.custos[4:6, 0])

    @property
    def custos_hidricos(self) -> float:
        """
        Parcela do custo de operação devido a violações de restrições
        hídricas, EVmin, VZmin, GHmin, GHmin Usina e outros usos
        da água, em MMR$.

        **Retorna**

        `float`
        """
        return (np.sum(self.custos[6:9, 0]) +
                np.sum(self.custos[12:14, 0]))

    @property
    def outros_custos(self) -> float:
        """
        Parcela do custo de operação devido a violações e penalidades
        de intercâmbio, intercâmbio mínimo, vertimento, vertimento
        fio d'água não turbinável, excesso de energia, emissão de GEE
        e retirada, em MMR$.

        **Retorna**

        `float`
        """
        return (np.sum(self.custos[2:4, 0]) +
                np.sum(self.custos[9:12, 0]) +
                np.sum(self.custos[14:, 0]))

    @property
    def custo_total(self) -> float:
        """
        Custo total de operação, em MMR$.

        **Retorna**

        `float`
        """
        return np.sum(self.custos[:, 0])


class PMO(Arquivo):
    """
    Armazena os dados de entrada do NEWAVE referentes ao
    acompanhamento do programa.

    Esta classe lida com as informações de entrada fornecidas ao
    NEWAVE e reproduzidas no `pmo.dat`, bem como as saídas finais
    da execução: custos de operação, energias, déficit, etc.

    Em versões futuras, esta classe pode passar a ler os dados
    de execução intermediárias do programa.

    """
    def __init__(self,
                 dados: DadosArquivo):
        super().__init__(dados)

        self.__eco_dger = self.__por_tipo(BlocoEcoDgerPMO)
        self.__convergencia = self.__por_tipo(BlocoConvergenciaPMO)
        self.__configs_exp = self.__por_tipo(BlocoConfiguracoesExpansaoPMO)
        self.__mars = self.__por_tipo(BlocoMARSPMO)
        self.__risco_deficit = self.__por_tipo(BlocoRiscoDeficitENSPMO)
        self.__custos = self.__por_tipo(BlocoCustoOperacaoPMO)

    def __por_tipo(self, tipo: Type[Bloco]) -> List[Bloco]:
        return [b for b in self._blocos if isinstance(b, tipo)]

    @classmethod
    def le_arquivo(cls,
                   diretorio: str,
                   nome_arquivo="pmo.dat") -> 'PMO':
        leitor = LeituraPMO(diretorio)
        r = leitor.le_arquivo(nome_arquivo)
        return cls(r)

    @property
    def configuracoes_entrada_reservatorio(self) -> np.ndarray:
        """
        Configurações do sistema em cada período devido a entrada
        de reservatórios e/ou potência de base.

        **Retorna**

        `np.ndarray`
        """
        return self.__configs_exp[0].dados

    @configuracoes_entrada_reservatorio.setter
    def configuracoes_entrada_reservatorio(self, configs: np.ndarray):
        self.__configs_exp[0].dados = configs

    @property
    def configuracoes_alteracao_potencia(self) -> np.ndarray:
        """
        Configurações do sistema em cada período devido a alterações
        de potência.

        **Retorna**

        `np.ndarray`
        """
        return self.__configs_exp[1].dados

    @configuracoes_alteracao_potencia.setter
    def configuracoes_alteracao_potencia(self, configs: np.ndarray):
        self.__configs_exp[1].dados = configs

    @property
    def configuracoes_qualquer_modificacao(self) -> np.ndarray:
        """
        Configurações do sistema em cada período devido a alterações
        de potência.

        **Retorna**

        `np.ndarray`
        """
        return self.__configs_exp[2].dados

    @configuracoes_qualquer_modificacao.setter
    def configuracoes_qualquer_modificacao(self, configs: np.ndarray):
        self.__configs_exp[2].dados = configs

    @property
    def retas_perdas_engolimento(self) -> Dict[str, np.ndarray]:
        """
        Retas ajustadas segundo o modelo MARS para corrigir a
        energia fio d'água com as perdas por engolimento máximo.

        OBS:
        Retorna apenas o modelo ajustado para a configuração do
        primeiro período de estudo (temporariamente).

        **Retorna**

        `np.ndarray`
        """
        retas: Dict[str, np.ndarray] = {}
        for i, r in enumerate(REES):
            retas[r] = self.__mars[0].dados[:, :, i]
        return retas

    @property
    def convergencia(self) -> pd.DataFrame:
        """
        Tabela de convergência da execução do NEWAVE.

        **Retorna**

        `pd.DataFrame`
        """
        return self.__convergencia[0].dados

    @property
    def risco_deficit_ens(self) -> pd.DataFrame:
        """
        Tabela de riscos de déficit e enegia não suprida (ENS).

        **Retorna**

        `pandas.DataFrame`
        """
        df = pd.DataFrame(self.__risco_deficit[0].dados,
                          columns=[
                                   "Ano",
                                   "Risco (%) SE",
                                   "EENS (MWMes) SE",
                                   "Risco (%) S",
                                   "EENS (MWMes) S",
                                   "Risco (%) NE",
                                   "EENS (MWMes) NE",
                                   "Risco (%) N",
                                   "EENS (MWMes) N"
                                  ])
        return df

    @property
    def custo_operacao_series_simuladas(self) -> pd.DataFrame:
        """
        Tabela de custos de operação categorizados para as
        séries simuladas.

        **Retorna**

        `pandas.DataFrame`
        """
        df = pd.DataFrame(self.__custos[0].dados,
                          columns=[
                                   "Valor Esperado",
                                   "Desvio Padrão do VE",
                                   "(%)"
                                  ])
        indices = pd.Series(BlocoCustoOperacaoPMO.componentes_custo)
        df.set_index(indices,
                     inplace=True)
        return df

    @property
    def valor_esperado_periodo_estudo(self) -> pd.DataFrame:
        """
        Tabela de custos de operação esperados para o período
        de estudo.

        **Retorna**

        `pandas.DataFrame`
        """
        df = pd.DataFrame(self.__custos[1].dados,
                          columns=[
                                   "Valor Esperado",
                                   "Desvio Padrão do VE",
                                   "(%)"
                                  ])
        indices = pd.Series(BlocoCustoOperacaoPMO.componentes_custo)
        df.set_index(indices,
                     inplace=True)
        return df

    @property
    def custo_operacao_referenciado_primeiro_mes(self) -> pd.DataFrame:
        """
        Tabela de custos de operação esperados para o período
        de estudo, referenciados ao primeiro mês.

        **Retorna**

        `pandas.DataFrame`
        """
        df = pd.DataFrame(self.__custos[2].dados,
                          columns=[
                                   "Valor Esperado",
                                   "Desvio Padrão do VE",
                                   "(%)"
                                  ])
        indices = pd.Series(BlocoCustoOperacaoPMO.componentes_custo)
        df.set_index(indices,
                     inplace=True)
        return df
